from .base import Backend


class GalaBackend(Backend):
    def __imports__():
        from gala import dynamics, potential

    def ORBIT_TYPE(self):
        return self.dynamics.orbit.Orbit

    def _compute_orbit(self, skycoord, pot, dt, steps):
        pos = [skycoord.x.value, skycoord.y.value, skycoord.z.value] * skycoord.x.unit
        vel = [
            skycoord.v_x.value,
            skycoord.v_y.value,
            skycoord.v_z.value,
        ] * skycoord.v_x.unit
        ics = self.dynamics.PhaseSpacePosition(pos, vel)
        orbit = self.potential.Hamiltonian(pot).integrate_orbit(
            ics, dt=dt, n_steps=steps - 1
        )
        return orbit

    def _extract_points(self, orbit):
        return orbit.data.T



from abc import abstractmethod, abstractproperty

import astropy.coordinates as coord
import astropy.units as u
from iext import ExtendImports


class Backend(ExtendImports):
    def __imports__():
        ...

    def __init__(self):
        self._args = None
        self._result = None

    @abstractproperty
    def ORBIT_TYPE(self):
        pass

    @abstractmethod
    def _compute_orbit(self, skycoord, pot, dt, steps):
        pass

    def compute_orbit(self, skycoord, pot, dt, steps):
        if not isinstance(skycoord, coord.SkyCoord):
            raise TypeError(
                "coord must be passed in as a astropy.coordinates.SkyCoord object."
                f"{skycoord} is not a astropy.coordinates.SkyCoord object."
            )
        skycoord.representation_type = "cartesian"
        if not (
            hasattr(skycoord, "v_x")
            and hasattr(skycoord, "v_y")
            and hasattr(skycoord, "v_z")
        ):
            raise TypeError(
                "coord must have 3-D velocity information to compute an orbit."
            )
        if not isinstance(dt, u.Quantity):
            raise TypeError(
                "dt must be passed in as a astropy.units.Quantity object."
                f"{dt} is not a astropy.coordinates.SkyCoord object."
            )
        self._args = skycoord, pot, dt, steps
        self._result = self._compute_orbit(skycoord, pot, dt, steps)
        return self._extract_points(self._result)

    @abstractmethod
    def _extract_points(self, orbit):
        pass

    def get_points(self):
        if not isinstance(self._result, self.ORBIT_TYPE):
            raise TypeError("No computed orbit to retrive points from.")
        return self._extract_points(self._result)



import inspect as _inspect
from collections.abc import Sequence as _Sequence

from .agama_backend import AgamaBackend
from .gala_backend import GalaBackend
from .galpy_backend import GalpyBackend


def get_backend_from(obj):
    if isinstance(obj, _Sequence) and len(obj) > 0:
        obj = obj[0]
    module = _inspect.getmodule(obj) or _inspect.getmodule(type(obj))
    if module is None:
        return None

    pkg, *_ = module.__name__.partition(".")
    if pkg == "agama":
        return AgamaBackend()
    elif pkg == "gala":
        return GalaBackend()
    elif pkg == "galpy":
        return GalpyBackend()
    return None



import astropy.coordinates as coord
import numpy as np

from .base import Backend


class GalpyBackend(Backend):
    def __imports__():
        from galpy import orbit

    def ORBIT_TYPE(self):
        return self.orbit.Orbit

    def _compute_orbit(self, skycoord, pot, dt, steps):
        orbit = self.orbit.Orbit(skycoord)
        t = np.arange(steps) * dt
        orbit.integrate(t, pot)
        return orbit

    def _extract_points(self, orbit):
        skycoord, pot, dt, steps = self._args
        t = np.arange(steps) * dt
        return coord.representation.CartesianRepresentation(
            orbit.x(t), orbit.y(t), orbit.z(t)
        )
